#| ======================================================================

 David Greve, David Hardin, Matthew Wilding

 Created : 3/23/1999

 ---

 This file illustrates three key issues encountered during the
 development of an Executable Formal Model (EFM) of an industrial
 scale microprocessor.  These issues are :

 1) A case for read-time evaluation of forms

 2) The need for a hybrid form that has properties of both functions
    and macros.

 3) A case for the support of 32-bit operations

 All of the issues presented herein were and continue to be real
 issues in the development of executable formal models for use
 in industrial applications.

====================================================================== |#

:ubt! 1

; (First, take care of a GCL optimization omission)
:q
(load "patch.lisp")
(lp)

(ld "ihs-helper.lisp")

:comp t

; Mark this point for future :ubt's
(deflabel ihs-loaded)

#| ======================================================================

; +++               #. (read time form evaluation)

=========================================================================

  - Common lisp provides both COND and CASE

    o multiple case conditional statements

  - GCL compiles CASE forms into efficient switch statements

    o not so for cond forms!

  - CASE forms do not evaluate their keys.

  - In complex systems, symbolic control representation is essential

    o It is common practice to use #. to force evaluation of runtime
      static symbolic keys within CASE forms

(defmacro KEY1 () 3)
(defmacro KEY2 () 4)
(defmacro KEY3 () 5)

(defun foo (KEYFORM)
  (case KEYFORM
	((#.(KEY1)) 1)
	((#.(KEY2)) 2)
	((#.(KEY3)) 3)
	(T 0))
  )

====================================================================== |#

; +++

; Example illustrating speed difference between CASE and COND

:ubt! ihs-loaded
(deflabel ihs-loaded)

(defmacro Int32 (x) `(the (signed-byte 32) ,x))

(defun alpha (x)

  (declare (type (signed-byte 32) x)
	   (xargs :verify-guards t))

  (case (Int32 (logand x #b1111))

	((#b0000) #b1111)
	((#b0001) #b1110)
	((#b0010) #b1101)
	((#b0011) #b1100)
	((#b0100) #b1011)
	((#b0101) #b1010)
	((#b0110) #b1001)
	((#b0111) #b1000)
	((#b1000) #b0111)
	((#b1001) #b0110)
	((#b1010) #b0101)
	((#b1011) #b0100)
	((#b1100) #b0011)
	((#b1101) #b0010)
	((#b1110) #b0001)
	((#b1111) #b0000)
	(t 0)

	)
  )

(defun beta (x)

  (declare (type (signed-byte 32) x)
	   (xargs :verify-guards t))

  (let ((y (logand x #b1111)))
    (declare (type (signed-byte 32) y))

    (cond 

	((equal y #b0000) #b1111)
	((equal y #b0001) #b1110)
	((equal y #b0010) #b1101)
	((equal y #b0011) #b1100)
	((equal y #b0100) #b1011)
	((equal y #b0101) #b1010)
	((equal y #b0110) #b1001)
	((equal y #b0111) #b1000)
	((equal y #b1000) #b0111)
	((equal y #b1001) #b0110)
	((equal y #b1010) #b0101)
	((equal y #b1011) #b0100)
	((equal y #b1100) #b0011)
	((equal y #b1101) #b0010)
	((equal y #b1110) #b0001)
	((equal y #b1111) #b0000)
	(t 0)

	)
    )
  )

(defun alpha-check (n b)

  (declare (type (signed-byte 32) n)
	   (type (signed-byte 32) b))

  (if (not (and (integerp n) (> n 0)))
      0
    (alpha-check (Int32 (1- n)) (alpha n))))

(defun beta-check (n b)

  (declare (type (signed-byte 32) n)
	   (type (signed-byte 32) b))

  (if (not (and (integerp n) (> n 0)))
      0
    (beta-check (Int32 (1- n)) (beta n))))

:comp t

; Run them in ACL2 to check for GC ..
;
(alpha-check #x7FFFFF 0)
( beta-check #x7FFFFF 0)

:q

; Run them in GCL to time them
;

; +++

(time (alpha-check #x7FFFFF 0))
; real time : 2.000 secs
; run time  : 1.550 secs

(time  (beta-check #x7FFFFF 0))
; real time : 3.000 secs
; run time  : 1.700 secs

(lp)

#| ======================================================================

; +++                  #. : A possible solution

=========================================================================

  - An ACL2 #. (even a somewhat restricted version thereof) would
    allow symbolic representation of contol words in case expressions

  - Alternately, a special form like CASE=, a case form that evaluates
    (runtime constant) key forms at read time and looks like a case
    form at compile time.

  - As it stands, we straddle the fence by creating a macro, CASE=, 
    that has different meanings depending on whether we are creating
    an EFM simulator (GCL) or doing proofs about the EFM (ACL2).

Following is an excerpt from the EFM model illustrating the use of a
case form that evaluates its keys, CASE=.  In ACL2, the CASE= macro
expands into a COND form.  In GCL, the CASE= macro evaluates it's keys
with EVAL and then creates a true CASE form.  It should be noted that
this sort of "dual meaning" is error prone and casts suspicion on our
claim of validating the ACL2 model through simulation.  In fact, there
was actually an error in the original definition of the GCL CASE=
which was discovered during the EFM model validation against the C
simulator. On the other hand, creating the EFM model without the use
of symbolic control fields would have been both error prone and
unacceptably time consuming.

; The ACL2 definition of CASE=
; ----------------------------

; +++

#-simulator
(defun key-list-expansion-of-case= (test keylist)
  (if (consp keylist)
      (cons `(= ,test ,(car keylist)) (key-list-expansion-of-case= test (cdr keylist)))
    nil))

#-simulator
(defun cond-expansion-of-case= (test list)
  (if (consp list)
      (let ((key (caar list)))
	(if (consp key)
	    (cons `((or ,@(key-list-expansion-of-case= test key)) ,(cadar list)) (cond-expansion-of-case= test (cdr list)))
            (if (equal key T)
		(cons `(t ,(cadar list)) (cond-expansion-of-case= test (cdr list)))
	      (cons `((= ,test ,key) ,(cadar list)) (cond-expansion-of-case= test (cdr list))))))
    nil))

#-simulator
(defmacro case= (test &rest rst)
  `(cond ,@(cond-expansion-of-case= test rst)))


; The GCL definition of CASE=
; ---------------------------

; +++

#+simulator
(defun eval-keys (keys)
  (if (consp keys)
      (if (consp (car keys))
          (cons (eval (car keys)) (eval-keys (cdr keys)))
	(eval keys))
      nil))
      
#+simulator
(defun case-keys (caselist)
  (if (consp caselist)
      (let ((line (first caselist)))
	(if (or (equal (car line) 't)
		(equal (car line) 'otherwise))
	    (list line)
	  (cons (cons (eval-keys (car line)) (cdr line))
		(case-keys (cdr caselist)))))
      nil))

#+simulator
(defmacro case= (test &rest rst)
  (let ((caselist (case-keys rst)))
  `(case ,test
     ,@caselist)))

; A fragment of the EFM model using case=
; ----------------------------------------

; +++

(defun ALU_OPERATE (st)
  (declare (xargs :guard (EFM-state-p st)
		  :guard-hints (("goal" :in-theory (set-difference-theories
						    (union-theories
						     (theory 'guard-proofs)
						     (current-theory :here))
						    (union-theories
						     (union-theories
						      (theory 'efm-state-updaters)
						      (theory 'efm-state-accessers))
						     '(efm-state-p ADDER ALULOG)))))))
  (BEGIN *state->state*

;  /* Perform ALU operation */
    (case= (or (Curr. ALUf) #x8) ; /* Force FN[3] = 1 to do pure ALU function */

          (((ALU_SUB_SR)) 
;    /* F = S + ~R + CIN */
	   (ADDER     (Curr. Sin)
		      (~ (Curr. Rin))
		      (Curr. Cin)
		       st))

          (((ALU_ADD)) 
;    /* F = S + R + CIN */
	   (ADDER  (Curr. Sin)
		   (Curr. Rin)  
		   (Curr. Cin)
		    st))

          (((ALU_XOR)
            (ALU_XOR_1))
;    /* F = S xor R */
	   (ALULOG 0 
		   (^ (Curr. Sin)
                      (Curr. Rin))
		   st))

          (((ALU_SUB_RS)) 
;    /* F = R + ~S + CIN */
	   (ADDER  (~ (Curr. Sin))
                      (Curr. Rin)
                      (Curr. CIN)
                   st))

          (((ALU_AND)) 
;    /* F = S and R */
	   (ALULOG 1
		   (& (Curr. Sin)
                      (Curr. Rin))
                   st))

          (((ALU_OR)) 
;    /* F = S or R */
	   (ALULOG 0
		   (or (Curr. Sin)
                       (Curr. Rin))
                   st))

          (((ALU_DIV)) 
;    /* if Z == 0 then F = S + R else F = S - R */
	   (if (true (== (Curr. Z) 0))
	       (ADDER (Curr. Sin)
		      (Curr. Rin)
		      (Curr. CIN)
		      st)
	   (else 
	       (ADDER (Curr. Sin)
		      (~ (Curr. Rin))
		      (Curr. CIN) 
		      st))))

          (t (Display_Illegal (ILLALUF) st))
    )))

=========================================================================

; +++                        DEFABREV

=========================================================================

+ Example illustrating "the" problems.

  - need declarations in order for GCL to make things fast.

  - declarations interfere with beta reduction of large,
    nested state updates.

    o proof w/guards

    o proof wout/guards

====================================================================== |#

:ubt! ihs-loaded
(deflabel ihs-loaded)

; +++

; Let's define our standard type "wrapper" macro as we would like to
; see them defined.  When we wrap functions in these macros, GCL is
; able to generate optimized (32-bit) code.
;

(defmacro Int32 (x) 
  `(the (signed-byte 32) ,x))

; Now load in the body of the function definitions and verify their
; guards.  Note that it takes around 160 seconds to load and verify
; the guards associated with "STEP-ST".
;

(ld "the-body.lisp")

; Time:  167.79 seconds (prove: 137.14, print: 23.00, other: 7.65)
;  STEP-ST

; +++

; At this point, we would like to load a proof about STEP-ST.
; However, this proof never seems to complete :(.  

;(ld "the-proof.lisp")

; Time:  ??? seconds

; This is especially troubling because it could be argued that, to
; circumvent the "THE" problem one could simply create two models of a
; function, one with wrappers and one without, and prove that they are
; equivalent.  However, aside from the fact that we don't want to have
; to create and maintain two versions of every function we write, that
; is essentially the technique being attempted in this example!
; However, because of massive case splitting and unsimplified lambda
; expressions, we are unable to complete even that simple proof.
;

; OK .. so let's try it a different way to illustrate our point.

:ubt! ihs-loaded
(deflabel ihs-loaded)

; +++

; If it were possible to treat certain macros like functions, we could
; do things in a different way.  Here we define the wrapper macro as a
; function .. emulating the type of behavior we would like to see with
; DEFABREV.  We then prove a lemma about the wrapper function and then
; disable it.
;

(defun Int32 (x) 
  (declare (type (signed-byte 32) x))
  (the (signed-byte 32) x))

(defthm satsifies-Int32
  (equal (Int32 x) x))

(in-theory (disable Int32))

; +++

; Now let's load the function definitions again.  Note that we still
; have the same proof obligations as before, but loading and guard
; checking "STEP-ST" now takes about 1/4 as long as in the previous
; example.
;

(ld "the-body.lisp")

; Time:  41.98 seconds (prove: 26.03, print: 7.91, other: 8.04)
;  STEP-ST

; More importantly, however, we are now able to load and prove a LEMMA
; about "STEP-ST" in rather short order.
;

(ld "the-proof.lisp")

; Time:  2.84 seconds (prove: 2.33, print: 0.43, other: 0.08)
;  STEP-EQUAL

; +++

; The definition of "STEP-ST" is similar in complexity to many of the
; functions used in the EFM model .. but there are several EFM
; functions that are in fact *more* complex than "STEP-ST".  We
; experienced first hand all of the problems illustrated in this
; example in spades with the EFM model.  As with the solution to the
; #. problem discussed above, our solution to this problem was to
; compile the EFM model in two different modes.  When we are building
; a simulator, the "wrapper" macros expand into type declarations to
; guide GCL in compiling the model efficiently.  However, when we
; build the formal model of the EFM, the macro definitions simply
; return their argument unchanged.  This allows us to do proofs about
; the EFM model.  Actually, we originally loaded the model with
; complete type wrappers into ACL2 to verify the guards.  This, by the
; way, was very painful and time consuming.  We then modified the
; wrapper definitions to allow proofs about the EFM model.

#| ======================================================================

; +++                        DEFABREV - IF

=========================================================================

+ Example illustrating "if" demotion.

  - case split on data

    o May be bad modeling style, but ..

    o May not really have a choice

====================================================================== |#

:ubt! ihs-loaded
(deflabel ihs-loaded)

(defmacro Int32 (x) `(the (signed-byte 32) ,x))

(defstobj st
  (CNTL :type (signed-byte 6) :initially 3)
  (DATA :type (signed-byte 6) :initially 4)
  )

; +++

; Using the single threaded object defined in "the-body", we now define a
; silly new function, FLIP-ST

(defun flip-st (st)

  (declare (xargs :stobjs (st)
		  :guard (stp st)))

  (let ((val (DATA st)))

    (if (< 0 val  )
	(setDATA -1 st)
      (setDATA 1 st)))

  )

; +++

; Symbolic simulation of flip-st performs a case split on a DATA
; element of the state .. this is not desirable since the DATA values
; during symbolic simulation will not be defined and the case splits
; (in general) would lead to exponential behavior.

(defthm flip-st-symbolic-simulation
  (equal (flip-st st) st))


;Subgoal 2
;(IMPLIES (< 0 (NTH 1 ST))
;         (EQUAL (SET-NTH 1 -1 ST) ST)).
;
;Name the formula above *1.
;
;Subgoal 1
;(IMPLIES (<= (NTH 1 ST) 0)
;         (EQUAL (SET-NTH 1 1 ST) ST)).


; +++

; Let's try to fix this by demoting the data dependant if's.  This
; isn't easy in ACL2 because it normally performs IF lifting and case
; splitting on ifs.  No problem, let's just define a function that
; "acts" like if ..

(defun my-if (a b c)
  (if a b c))

; Prove some nice lemmas about it, like :

(defthm my-if-demotion
  (equal (my-if a (set-nth i v st) (set-nth i w st))
	 (set-nth i (my-if a v w) st)))

; +++

; And then shut it off.

(in-theory (disable my-if))

; +++

; This rule should give us the desired behavior of my-if over STOBJ
; state updates.  Let's go back and try to apply it ..

#| ======================================================================

(defun alt-flip-st (st)

  (declare (xargs :stobjs (st)
		  :guard (stp st)))

  (let ((val (DATA st)))

    (my-if (< 0 val)
	(setDATA -1 st)
      (setDATA 1 st)))

  )

====================================================================== |#

; +++

; Oops .. this violates single threadedness!! Why?  The arguments to a
; function are all evaluated prior to execution of that function.  The
; STOBJ single threadedness rules detect this problem and won't accept
; the above function definition.  So that solution won't work!

; Logicly we aren't completely without hope, of course.  It should
; usually be possible to demote if's by hand by re-writing offensive
; terms.  This isn't what we want because we may not have had control over
; the original model and we don't want to maintain two different models.

; Nonetheless let's give it a shot.  Now, since we can't use a generic
; "my_if" anyway (state cannot be an argument), let's optimize it for
; fixnums ..

(defun my-if* (a b c)
  (declare (type (signed-byte 32) b)
	   (type (signed-byte 32) c))

  (the (signed-byte 32) (if a b c)))

; +++

; Define a new version of FLIP_ST

(defun alt-flip-st (st)

  (declare (xargs :stobjs (st)
		  :guard (stp st)))

  (let ((val (DATA st)))

    (setDATA (my-if* (< 0 val) -1 1) st))

  )

; +++

; We can even prove that this new function is the same as the original.

(defthm alt-fix-is-fix
  (equal (alt-flip-st st) (flip-st st)))

; +++

; However, when it comes to execution, we now take a hit .. we have
; effectively increased the complexity of a simple control structure
; by adding a function call!  Not exactly strength reduction.

(defun iterate-flip-st (n st)

  (declare (type (signed-byte 32) n)
	   (xargs :stobjs (st)
		  :guard (stp st)))

  (if (not (and (integerp n) (> n 0)))
      st
    (let ((st (flip-st st)))
      (iterate-flip-st (Int32 (1- n)) st))))


(defun iterate-alt-flip-st (n st)

  (declare (type (signed-byte 32) n)
	   (xargs :stobjs (st)
		  :guard (stp st)))

  (if (not (and (integerp n) (> n 0)))
      st
    (let ((st (alt-flip-st st)))
      (iterate-alt-flip-st (Int32 (1- n)) st))))


:comp t

(iterate-flip-st #x7FFFFF st)
(iterate-alt-flip-st #x7FFFFF st)

:q

; +++

(time     (iterate-flip-st #x3FFFFF st))
; real time : 2.000 secs
; run time  : 2.730 secs

(time (iterate-alt-flip-st #x3FFFFF st))
; real time : 3.000 secs
; run time  : 3.330 secs

(lp)

; +++

; So what do we "need"? We need a symbol that acts like a function from
; the perspective of the ACL2 theorem prover but acts like a macro
; from the perspective of the Lisp compiler.

#| ======================================================================

; +++                 32-bit (fixnum) addition

  When we build models of hardware, we inevitably want to model
modular arithmetic.  At the same time, however, we want to use
ACL2/Common Lisp to build our models so that they are analyzable and
can execute efficiently.

====================================================================== |#

:ubt! ihs-loaded
(deflabel ihs-loaded)

; +++

; Here we define a function to perform addition.

(defun p32 (a b) 
  (declare (xargs :guard (and (acl2-numberp a) (acl2-numberp b))))
  (+ a b))

; And then define an iterator and compile the functions.

(defun repeat (x a b) 
  (declare (xargs :guard (and (acl2-numberp a) (acl2-numberp b))))
  (if (or (not (integerp x)) (<= x 0))
       a
     (let ((temp (p32 a b)))
       (repeat (1- x) temp b))))

:comp t

:q

; +++

; However, Common Lisp integers are, by nature, bignums.  So, unlike C
; where the results of arithmetic operations are automatically cast as
; 32-bit numbers, in Common Lisp we get the infinite-word-size result.
; Although such precision is sometimes handy, it is generally quite
; slow.

; +++

(time (repeat 1000000 #x7FFFFFF0 1))

; Results in :
;
; [SGC for 85 FIXNUM pages..(4687 writable)..(T=32).GC finished]
; [SGC for 85 FIXNUM pages..(4687 writable)..(T=34).GC finished]
; ...
; [SGC for 85 FIXNUM pages..(4687 writable)..(T=32).GC finished]
; [SGC for 85 FIXNUM pages..(4687 writable)..(T=33).GC finished]
; real time : 19.000 secs
; run time  : 18.110 secs
; 2148483632

; +++

; Common Lisp does, however, support declarations for "fixnums."
; Perhaps we can declare values so that the compiler knows enough to
; compile to a fast version.

(lp)

:ubt! ihs-loaded
(deflabel ihs-loaded)

; +++

; Define a type wrapper macro

(defmacro Int32 (x) `(the (signed-byte 32) ,x))

; +++

; and define a new version of p32 using type declarations

(defun p32 (x y)
   (declare 
     (xargs :verify-guards nil)
     (type (signed-byte 32) x y))
   (Int32 (+ (Int32 x) (Int32 y))))

(defun repeat (x a b) 
   (declare 
    (type (signed-byte 32) x a b)
    (xargs :verify-guards nil))
   (if (or (not (integerp (Int32 x))) (<= (Int32 x) (Int32 0)))
     (Int32 a)
     (let ((temp (p32 (Int32 a) (Int32 b))))
       (declare (type (signed-byte 32) temp))
       (repeat (Int32 (1- (Int32 x))) (Int32 temp) (Int32 b)))))

:comp t

:q

; +++

; This gets compiled by GCL to execute using the 32-bit addition
; provided by the underlying machine.  

(time (repeat 1000000 #x7FFFFFF0 1))
; real time : 0.000 secs
; run time  : 0.110 secs
; -2146483664

(lp)

; +++

; But, in ACL2, the guards are unsatisfiable, reflecting the simple
; fact that there are some 32-bit integers whose sum is not
; representable in 32-bits.  This is evidenced by the previous example
; when we note that, using bignums, the answer was 2148483632, not
; -2146483664.  What this means is that, in GCL, the type declaration
; was treated as an "unsafe" coercion.

; Nonetheless, the behavior of the coerced function is what we want!

; The problem is that we don't want "+" to be the common lisp "+", we
; want it to be the 32 bit modular plus that we have come to know and
; love in C.  We need to find a way to convince both GCL and ACL2 that
; we are performing 32-bit modular addition.  So, what if we simply
; modeled the "C" plus that we want in the logic of ACL2?

:ubt! ihs-loaded
(deflabel ihs-loaded)

(defmacro Int32 (x) `(the (signed-byte 32) ,x))

; +++

; Here we model 32-bit modular addition

(defun p32 (x y)
     (declare
       (xargs :guard-hints (("goal" :use (:instance signed-byte-p-logext (size1 32) (size 32) (i (+ x y)))
                                    :in-theory (union-theories (disable SIGNED-BYTE-P-LOGEXT) '(signed-byte-p)))))
       (type (signed-byte 32) x y))
     (Int32 (logext 32 (+ (Int32 x) (Int32 y)))))

(defun repeat (x a b) 
   (declare (type (signed-byte 32) x a b)
	    (xargs :guard-hints (("goal" :use (:instance signed-byte-p-logext (size1 32) (size 32) (i (+ a b)))
                                    :in-theory (union-theories (disable SIGNED-BYTE-P-LOGEXT) '(signed-byte-p))))))
   (if (or (not (integerp x)) (<= x 0))
     a
     (let ((temp (p32 (Int32 a) (Int32 b))))
       (declare (type (signed-byte 32) temp))
       (repeat (Int32 (1- x)) temp b))))

:comp t

:q

; +++

; So now we *can* satisfy our guard conditions.  However ..

(time (repeat 1000000 #x7FFFFFF0 1))

; Results in :
;
; [SGC for 85 FIXNUM pages..(4687 writable)..(T=32).GC finished]
; [SGC for 85 FIXNUM pages..(4687 writable)..(T=34).GC finished]
; ...
; [SGC for 85 FIXNUM pages..(4687 writable)..(T=32).GC finished]
; [SGC for 85 FIXNUM pages..(4687 writable)..(T=33).GC finished]
; real time : 216.000 secs
; run time  : 213.930 secs
; -2146483664

(lp)

; +++

; Whew!  Now it is *really* slow!  Of course, that was to be expected.
; Simply modeling the functionality will not give us the desired
; performance.
;
; However, if we are clever, we can now replace the function p32
; "under the hood" with its (assumed) equivalent "C" implementation
; using GCL compiler optimizations.  This gives us the desired C speed
; without violating the ACL2 interpretation of the common lisp "+"
; function.

:ubt! ihs-loaded
(deflabel ihs-loaded)

(defmacro Int32 (x) `(the (signed-byte 32) ,x))

; +++

; Re-do the p32 definition

(defun p32 (x y)
     (declare
       (xargs :guard-hints (("goal" :use (:instance signed-byte-p-logext (size1 32) (size 32) (i (+ x y)))
                                    :in-theory (union-theories (disable SIGNED-BYTE-P-LOGEXT) '(signed-byte-p)))))
       (type (signed-byte 32) x y))
     (Int32 (logext 32 (+ (Int32 x) (Int32 y)))))

(defun repeat (x a b) 
   (declare (type (signed-byte 32) x a b)
	    (xargs :guard-hints (("goal" :use (:instance signed-byte-p-logext (size1 32) (size 32) (i (+ a b)))
                                    :in-theory (union-theories (disable SIGNED-BYTE-P-LOGEXT) '(signed-byte-p))))))
   (if (or (not (integerp x)) (<= x 0))
     a
     (let ((temp (p32 (Int32 a) (Int32 b))))
       (declare (type (signed-byte 32) temp))
       (repeat (Int32 (1- x)) temp b))))

; +++

; But, under the hood ..

:q
(in-package "COMPILER")
(push '((fixnum fixnum) fixnum 0 "((#0) + (#1))")
  (get 'ACL2::p32 'inline-always))
(in-package "ACL2")
(lp)

; +++

; Note that, because of the under the hood GCL compiler optimization,
; the above compilation did not require the compilation of the ACL2
; definition of p32 in order to run fast. (heh!)

:comp repeat

(repeat 1000000 #x7FFFFFF0 1)

:q

; +++

(time (repeat 1000000 #x7FFFFFF0 1))

; Results in :
;
; real time : 0.000 secs
; run time  : 0.010 secs
; -2146483664

(lp)


; +++

; Of course, we have now made the assumption that p32 correctly models
; the C implementation of + .  Is this the case?  This is a issue that
; we would much rather have addressed by the experts !!

